/*
 * Copyright (C) 2010-2021 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#if ENABLE(GPU_PROCESS)
#include "GPUConnectionToWebProcess.h"

#include "Decoder.h"
#include "GPUConnectionToWebProcessMessages.h"
#if ENABLE(WEBGL)
#include "GraphicsContextGLIdentifier.h"
#endif
#include "HandleMessage.h"
#include "IPCSemaphore.h"
#include "MediaOverridesForTesting.h"
#include "RemoteAudioHardwareListenerIdentifier.h"
#if USE(AUDIO_SESSION)
#include "RemoteAudioSessionConfiguration.h"
#endif
#include "RemoteRemoteCommandListenerIdentifier.h"
#include "RenderingBackendIdentifier.h"
#if ENABLE(WEBGL)
#include "StreamConnectionBuffer.h"
#endif
#include "WebCoreArgumentCoders.h"
#if ENABLE(WEBGL)
#include <WebCore/GraphicsContextGLAttributes.h>
#endif
#include <WebCore/NowPlayingInfo.h>

namespace Messages {

namespace GPUConnectionToWebProcess {

#if USE(AUDIO_SESSION)

void EnsureAudioSession::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, const WebKit::RemoteAudioSessionConfiguration& configuration)
{
    *encoder << configuration;
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

} // namespace GPUConnectionToWebProcess

} // namespace Messages

namespace WebKit {

void GPUConnectionToWebProcess::didReceiveMessage(IPC::Connection& connection, IPC::Decoder& decoder)
{
    auto protectedThis = makeRef(*this);
    if (decoder.messageName() == Messages::GPUConnectionToWebProcess::CreateRenderingBackend::name()) {
        IPC::handleMessage<Messages::GPUConnectionToWebProcess::CreateRenderingBackend>(decoder, this, &GPUConnectionToWebProcess::createRenderingBackend);
        return;
    }
    if (decoder.messageName() == Messages::GPUConnectionToWebProcess::ReleaseRenderingBackend::name()) {
        IPC::handleMessage<Messages::GPUConnectionToWebProcess::ReleaseRenderingBackend>(decoder, this, &GPUConnectionToWebProcess::releaseRenderingBackend);
        return;
    }
#if ENABLE(WEBGL)
    if (decoder.messageName() == Messages::GPUConnectionToWebProcess::CreateGraphicsContextGL::name()) {
        IPC::handleMessage<Messages::GPUConnectionToWebProcess::CreateGraphicsContextGL>(decoder, this, &GPUConnectionToWebProcess::createGraphicsContextGL);
        return;
    }
#endif
#if ENABLE(WEBGL)
    if (decoder.messageName() == Messages::GPUConnectionToWebProcess::ReleaseGraphicsContextGL::name()) {
        IPC::handleMessage<Messages::GPUConnectionToWebProcess::ReleaseGraphicsContextGL>(decoder, this, &GPUConnectionToWebProcess::releaseGraphicsContextGL);
        return;
    }
#endif
    if (decoder.messageName() == Messages::GPUConnectionToWebProcess::ClearNowPlayingInfo::name()) {
        IPC::handleMessage<Messages::GPUConnectionToWebProcess::ClearNowPlayingInfo>(decoder, this, &GPUConnectionToWebProcess::clearNowPlayingInfo);
        return;
    }
    if (decoder.messageName() == Messages::GPUConnectionToWebProcess::SetNowPlayingInfo::name()) {
        IPC::handleMessage<Messages::GPUConnectionToWebProcess::SetNowPlayingInfo>(decoder, this, &GPUConnectionToWebProcess::setNowPlayingInfo);
        return;
    }
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::GPUConnectionToWebProcess::EnsureMediaSessionHelper::name()) {
        IPC::handleMessage<Messages::GPUConnectionToWebProcess::EnsureMediaSessionHelper>(decoder, this, &GPUConnectionToWebProcess::ensureMediaSessionHelper);
        return;
    }
#endif
#if ENABLE(VP9)
    if (decoder.messageName() == Messages::GPUConnectionToWebProcess::EnableVP9Decoders::name()) {
        IPC::handleMessage<Messages::GPUConnectionToWebProcess::EnableVP9Decoders>(decoder, this, &GPUConnectionToWebProcess::enableVP9Decoders);
        return;
    }
#endif
    if (decoder.messageName() == Messages::GPUConnectionToWebProcess::SetMediaOverridesForTesting::name()) {
        IPC::handleMessage<Messages::GPUConnectionToWebProcess::SetMediaOverridesForTesting>(decoder, this, &GPUConnectionToWebProcess::setMediaOverridesForTesting);
        return;
    }
    if (decoder.messageName() == Messages::GPUConnectionToWebProcess::CreateAudioHardwareListener::name()) {
        IPC::handleMessage<Messages::GPUConnectionToWebProcess::CreateAudioHardwareListener>(decoder, this, &GPUConnectionToWebProcess::createAudioHardwareListener);
        return;
    }
    if (decoder.messageName() == Messages::GPUConnectionToWebProcess::ReleaseAudioHardwareListener::name()) {
        IPC::handleMessage<Messages::GPUConnectionToWebProcess::ReleaseAudioHardwareListener>(decoder, this, &GPUConnectionToWebProcess::releaseAudioHardwareListener);
        return;
    }
    if (decoder.messageName() == Messages::GPUConnectionToWebProcess::CreateRemoteCommandListener::name()) {
        IPC::handleMessage<Messages::GPUConnectionToWebProcess::CreateRemoteCommandListener>(decoder, this, &GPUConnectionToWebProcess::createRemoteCommandListener);
        return;
    }
    if (decoder.messageName() == Messages::GPUConnectionToWebProcess::ReleaseRemoteCommandListener::name()) {
        IPC::handleMessage<Messages::GPUConnectionToWebProcess::ReleaseRemoteCommandListener>(decoder, this, &GPUConnectionToWebProcess::releaseRemoteCommandListener);
        return;
    }
    if (dispatchMessage(connection, decoder))
        return;
    UNUSED_PARAM(connection);
    UNUSED_PARAM(decoder);
    ASSERT_NOT_REACHED();
}

void GPUConnectionToWebProcess::didReceiveSyncMessage(IPC::Connection& connection, IPC::Decoder& decoder, std::unique_ptr<IPC::Encoder>& replyEncoder)
{
    auto protectedThis = makeRef(*this);
#if USE(AUDIO_SESSION)
    if (decoder.messageName() == Messages::GPUConnectionToWebProcess::EnsureAudioSession::name()) {
        IPC::handleMessageSynchronous<Messages::GPUConnectionToWebProcess::EnsureAudioSession>(connection, decoder, replyEncoder, this, &GPUConnectionToWebProcess::ensureAudioSession);
        return;
    }
#endif
    if (dispatchSyncMessage(connection, decoder, replyEncoder))
        return;
    UNUSED_PARAM(connection);
    UNUSED_PARAM(decoder);
    UNUSED_PARAM(replyEncoder);
    ASSERT_NOT_REACHED();
}

} // namespace WebKit

#endif // ENABLE(GPU_PROCESS)
