/*
 * Copyright (C) 2010-2021 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#if ENABLE(GPU_PROCESS)
#include "GPUProcess.h"

#include "ArgumentCoders.h"
#include "Attachment.h"
#include "Decoder.h"
#include "GPUProcessConnectionParameters.h"
#include "GPUProcessCreationParameters.h"
#include "GPUProcessMessages.h"
#include "GPUProcessSessionParameters.h"
#include "HandleMessage.h"
#include "WebCoreArgumentCoders.h"
#if ENABLE(MEDIA_STREAM)
#include <WebCore/MockMediaDevice.h>
#endif
#include <WebCore/ProcessIdentifier.h>
#include <pal/SessionID.h>
#include <wtf/Optional.h>
#if ENABLE(MEDIA_STREAM)
#include <wtf/text/WTFString.h>
#endif

namespace Messages {

namespace GPUProcess {

void CreateGPUConnectionToWebProcess::callReply(IPC::Decoder& decoder, CompletionHandler<void(Optional<IPC::Attachment>&&)>&& completionHandler)
{
    Optional<Optional<IPC::Attachment>> connectionIdentifier;
    decoder >> connectionIdentifier;
    if (!connectionIdentifier) {
        ASSERT_NOT_REACHED();
        cancelReply(WTFMove(completionHandler));
        return;
    }
    completionHandler(WTFMove(*connectionIdentifier));
}

void CreateGPUConnectionToWebProcess::cancelReply(CompletionHandler<void(Optional<IPC::Attachment>&&)>&& completionHandler)
{
    completionHandler(IPC::AsyncReplyError<Optional<IPC::Attachment>>::create());
}

void CreateGPUConnectionToWebProcess::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, const Optional<IPC::Attachment>& connectionIdentifier)
{
    *encoder << connectionIdentifier;
    connection.sendSyncReply(WTFMove(encoder));
}

void PrepareToSuspend::callReply(IPC::Decoder& decoder, CompletionHandler<void()>&& completionHandler)
{
    completionHandler();
}

void PrepareToSuspend::cancelReply(CompletionHandler<void()>&& completionHandler)
{
    completionHandler();
}

void PrepareToSuspend::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection)
{
    connection.sendSyncReply(WTFMove(encoder));
}

#if ENABLE(MEDIA_STREAM)

void UpdateCaptureAccess::callReply(IPC::Decoder& decoder, CompletionHandler<void()>&& completionHandler)
{
    completionHandler();
}

void UpdateCaptureAccess::cancelReply(CompletionHandler<void()>&& completionHandler)
{
    completionHandler();
}

void UpdateCaptureAccess::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection)
{
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

} // namespace GPUProcess

} // namespace Messages

namespace WebKit {

void GPUProcess::didReceiveMessage(IPC::Connection& connection, IPC::Decoder& decoder)
{
    auto protectedThis = makeRef(*this);
    if (decoder.messageName() == Messages::GPUProcess::InitializeGPUProcess::name()) {
        IPC::handleMessage<Messages::GPUProcess::InitializeGPUProcess>(decoder, this, &GPUProcess::initializeGPUProcess);
        return;
    }
    if (decoder.messageName() == Messages::GPUProcess::CreateGPUConnectionToWebProcess::name()) {
        IPC::handleMessageAsync<Messages::GPUProcess::CreateGPUConnectionToWebProcess>(connection, decoder, this, &GPUProcess::createGPUConnectionToWebProcess);
        return;
    }
    if (decoder.messageName() == Messages::GPUProcess::ProcessDidTransitionToForeground::name()) {
        IPC::handleMessage<Messages::GPUProcess::ProcessDidTransitionToForeground>(decoder, this, &GPUProcess::processDidTransitionToForeground);
        return;
    }
    if (decoder.messageName() == Messages::GPUProcess::ProcessDidTransitionToBackground::name()) {
        IPC::handleMessage<Messages::GPUProcess::ProcessDidTransitionToBackground>(decoder, this, &GPUProcess::processDidTransitionToBackground);
        return;
    }
    if (decoder.messageName() == Messages::GPUProcess::PrepareToSuspend::name()) {
        IPC::handleMessageAsync<Messages::GPUProcess::PrepareToSuspend>(connection, decoder, this, &GPUProcess::prepareToSuspend);
        return;
    }
    if (decoder.messageName() == Messages::GPUProcess::ProcessDidResume::name()) {
        IPC::handleMessage<Messages::GPUProcess::ProcessDidResume>(decoder, this, &GPUProcess::processDidResume);
        return;
    }
    if (decoder.messageName() == Messages::GPUProcess::AddSession::name()) {
        IPC::handleMessage<Messages::GPUProcess::AddSession>(decoder, this, &GPUProcess::addSession);
        return;
    }
    if (decoder.messageName() == Messages::GPUProcess::RemoveSession::name()) {
        IPC::handleMessage<Messages::GPUProcess::RemoveSession>(decoder, this, &GPUProcess::removeSession);
        return;
    }
#if ENABLE(MEDIA_STREAM)
    if (decoder.messageName() == Messages::GPUProcess::SetMockCaptureDevicesEnabled::name()) {
        IPC::handleMessage<Messages::GPUProcess::SetMockCaptureDevicesEnabled>(decoder, this, &GPUProcess::setMockCaptureDevicesEnabled);
        return;
    }
#endif
#if ENABLE(MEDIA_STREAM)
    if (decoder.messageName() == Messages::GPUProcess::SetOrientationForMediaCapture::name()) {
        IPC::handleMessage<Messages::GPUProcess::SetOrientationForMediaCapture>(decoder, this, &GPUProcess::setOrientationForMediaCapture);
        return;
    }
#endif
#if ENABLE(MEDIA_STREAM)
    if (decoder.messageName() == Messages::GPUProcess::UpdateCaptureAccess::name()) {
        IPC::handleMessageAsync<Messages::GPUProcess::UpdateCaptureAccess>(connection, decoder, this, &GPUProcess::updateCaptureAccess);
        return;
    }
#endif
#if ENABLE(MEDIA_STREAM)
    if (decoder.messageName() == Messages::GPUProcess::AddMockMediaDevice::name()) {
        IPC::handleMessage<Messages::GPUProcess::AddMockMediaDevice>(decoder, this, &GPUProcess::addMockMediaDevice);
        return;
    }
#endif
#if ENABLE(MEDIA_STREAM)
    if (decoder.messageName() == Messages::GPUProcess::ClearMockMediaDevices::name()) {
        IPC::handleMessage<Messages::GPUProcess::ClearMockMediaDevices>(decoder, this, &GPUProcess::clearMockMediaDevices);
        return;
    }
#endif
#if ENABLE(MEDIA_STREAM)
    if (decoder.messageName() == Messages::GPUProcess::RemoveMockMediaDevice::name()) {
        IPC::handleMessage<Messages::GPUProcess::RemoveMockMediaDevice>(decoder, this, &GPUProcess::removeMockMediaDevice);
        return;
    }
#endif
#if ENABLE(MEDIA_STREAM)
    if (decoder.messageName() == Messages::GPUProcess::ResetMockMediaDevices::name()) {
        IPC::handleMessage<Messages::GPUProcess::ResetMockMediaDevices>(decoder, this, &GPUProcess::resetMockMediaDevices);
        return;
    }
#endif
    UNUSED_PARAM(connection);
    UNUSED_PARAM(decoder);
    ASSERT_NOT_REACHED();
}

} // namespace WebKit

#endif // ENABLE(GPU_PROCESS)
