/*
 * Copyright (C) 2010-2021 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#if ENABLE(GPU_PROCESS) && ENABLE(WEB_AUDIO)
#include "RemoteAudioDestinationManager.h"

#include "ArgumentCoders.h"
#include "Decoder.h"
#include "HandleMessage.h"
#include "IPCSemaphore.h"
#include "RemoteAudioDestinationIdentifier.h"
#include "RemoteAudioDestinationManagerMessages.h"
#if PLATFORM(COCOA)
#include "SharedMemory.h"
#endif
#if PLATFORM(COCOA)
#include "WebCoreArgumentCoders.h"
#endif
#if PLATFORM(COCOA)
#include <WebCore/CAAudioStreamDescription.h>
#endif
#include <wtf/text/WTFString.h>

namespace Messages {

namespace RemoteAudioDestinationManager {

void CreateAudioDestination::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, const WebKit::RemoteAudioDestinationIdentifier& identifier)
{
    *encoder << identifier;
    connection.sendSyncReply(WTFMove(encoder));
}

void DeleteAudioDestination::callReply(IPC::Decoder& decoder, CompletionHandler<void()>&& completionHandler)
{
    completionHandler();
}

void DeleteAudioDestination::cancelReply(CompletionHandler<void()>&& completionHandler)
{
    completionHandler();
}

void DeleteAudioDestination::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection)
{
    connection.sendSyncReply(WTFMove(encoder));
}

void StartAudioDestination::callReply(IPC::Decoder& decoder, CompletionHandler<void(bool&&)>&& completionHandler)
{
    Optional<bool> isPlaying;
    decoder >> isPlaying;
    if (!isPlaying) {
        ASSERT_NOT_REACHED();
        cancelReply(WTFMove(completionHandler));
        return;
    }
    completionHandler(WTFMove(*isPlaying));
}

void StartAudioDestination::cancelReply(CompletionHandler<void(bool&&)>&& completionHandler)
{
    completionHandler(IPC::AsyncReplyError<bool>::create());
}

void StartAudioDestination::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, bool isPlaying)
{
    *encoder << isPlaying;
    connection.sendSyncReply(WTFMove(encoder));
}

void StopAudioDestination::callReply(IPC::Decoder& decoder, CompletionHandler<void(bool&&)>&& completionHandler)
{
    Optional<bool> isPlaying;
    decoder >> isPlaying;
    if (!isPlaying) {
        ASSERT_NOT_REACHED();
        cancelReply(WTFMove(completionHandler));
        return;
    }
    completionHandler(WTFMove(*isPlaying));
}

void StopAudioDestination::cancelReply(CompletionHandler<void(bool&&)>&& completionHandler)
{
    completionHandler(IPC::AsyncReplyError<bool>::create());
}

void StopAudioDestination::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, bool isPlaying)
{
    *encoder << isPlaying;
    connection.sendSyncReply(WTFMove(encoder));
}

} // namespace RemoteAudioDestinationManager

} // namespace Messages

namespace WebKit {

void RemoteAudioDestinationManager::didReceiveMessage(IPC::Connection& connection, IPC::Decoder& decoder)
{
    if (decoder.messageName() == Messages::RemoteAudioDestinationManager::DeleteAudioDestination::name()) {
        IPC::handleMessageAsync<Messages::RemoteAudioDestinationManager::DeleteAudioDestination>(connection, decoder, this, &RemoteAudioDestinationManager::deleteAudioDestination);
        return;
    }
    if (decoder.messageName() == Messages::RemoteAudioDestinationManager::StartAudioDestination::name()) {
        IPC::handleMessageAsync<Messages::RemoteAudioDestinationManager::StartAudioDestination>(connection, decoder, this, &RemoteAudioDestinationManager::startAudioDestination);
        return;
    }
    if (decoder.messageName() == Messages::RemoteAudioDestinationManager::StopAudioDestination::name()) {
        IPC::handleMessageAsync<Messages::RemoteAudioDestinationManager::StopAudioDestination>(connection, decoder, this, &RemoteAudioDestinationManager::stopAudioDestination);
        return;
    }
#if PLATFORM(COCOA)
    if (decoder.messageName() == Messages::RemoteAudioDestinationManager::AudioSamplesStorageChanged::name()) {
        IPC::handleMessage<Messages::RemoteAudioDestinationManager::AudioSamplesStorageChanged>(decoder, this, &RemoteAudioDestinationManager::audioSamplesStorageChanged);
        return;
    }
#endif
    UNUSED_PARAM(connection);
    UNUSED_PARAM(decoder);
    ASSERT_NOT_REACHED();
}

void RemoteAudioDestinationManager::didReceiveSyncMessage(IPC::Connection& connection, IPC::Decoder& decoder, std::unique_ptr<IPC::Encoder>& replyEncoder)
{
    if (decoder.messageName() == Messages::RemoteAudioDestinationManager::CreateAudioDestination::name()) {
        IPC::handleMessageSynchronous<Messages::RemoteAudioDestinationManager::CreateAudioDestination>(connection, decoder, replyEncoder, this, &RemoteAudioDestinationManager::createAudioDestination);
        return;
    }
    UNUSED_PARAM(connection);
    UNUSED_PARAM(decoder);
    UNUSED_PARAM(replyEncoder);
    ASSERT_NOT_REACHED();
}

} // namespace WebKit

#endif // ENABLE(GPU_PROCESS) && ENABLE(WEB_AUDIO)
