/*
 * Copyright (C) 2010-2021 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#if ENABLE(GPU_PROCESS) && ENABLE(ENCRYPTED_MEDIA)
#include "RemoteCDMInstanceProxy.h"

#include "ArgumentCoders.h"
#include "Decoder.h"
#include "HandleMessage.h"
#include "RemoteCDMInstanceProxyMessages.h"
#include "RemoteCDMInstanceSessionIdentifier.h"
#include "SharedBufferCopy.h"
#include "WebCoreArgumentCoders.h"
#include <WebCore/CDMInstance.h>
#include <WebCore/CDMKeySystemConfiguration.h>
#include <wtf/text/WTFString.h>

namespace Messages {

namespace RemoteCDMInstanceProxy {

void CreateSession::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, const WebKit::RemoteCDMInstanceSessionIdentifier& identifier)
{
    *encoder << identifier;
    connection.sendSyncReply(WTFMove(encoder));
}

void InitializeWithConfiguration::callReply(IPC::Decoder& decoder, CompletionHandler<void(WebCore::CDMInstance::SuccessValue&&)>&& completionHandler)
{
    Optional<WebCore::CDMInstance::SuccessValue> success;
    decoder >> success;
    if (!success) {
        ASSERT_NOT_REACHED();
        cancelReply(WTFMove(completionHandler));
        return;
    }
    completionHandler(WTFMove(*success));
}

void InitializeWithConfiguration::cancelReply(CompletionHandler<void(WebCore::CDMInstance::SuccessValue&&)>&& completionHandler)
{
    completionHandler(IPC::AsyncReplyError<WebCore::CDMInstance::SuccessValue>::create());
}

void InitializeWithConfiguration::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, WebCore::CDMInstance::SuccessValue success)
{
    *encoder << success;
    connection.sendSyncReply(WTFMove(encoder));
}

void SetServerCertificate::callReply(IPC::Decoder& decoder, CompletionHandler<void(WebCore::CDMInstance::SuccessValue&&)>&& completionHandler)
{
    Optional<WebCore::CDMInstance::SuccessValue> success;
    decoder >> success;
    if (!success) {
        ASSERT_NOT_REACHED();
        cancelReply(WTFMove(completionHandler));
        return;
    }
    completionHandler(WTFMove(*success));
}

void SetServerCertificate::cancelReply(CompletionHandler<void(WebCore::CDMInstance::SuccessValue&&)>&& completionHandler)
{
    completionHandler(IPC::AsyncReplyError<WebCore::CDMInstance::SuccessValue>::create());
}

void SetServerCertificate::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, WebCore::CDMInstance::SuccessValue success)
{
    *encoder << success;
    connection.sendSyncReply(WTFMove(encoder));
}

} // namespace RemoteCDMInstanceProxy

} // namespace Messages

namespace WebKit {

void RemoteCDMInstanceProxy::didReceiveMessage(IPC::Connection& connection, IPC::Decoder& decoder)
{
    if (decoder.messageName() == Messages::RemoteCDMInstanceProxy::InitializeWithConfiguration::name()) {
        IPC::handleMessageAsync<Messages::RemoteCDMInstanceProxy::InitializeWithConfiguration>(connection, decoder, this, &RemoteCDMInstanceProxy::initializeWithConfiguration);
        return;
    }
    if (decoder.messageName() == Messages::RemoteCDMInstanceProxy::SetServerCertificate::name()) {
        IPC::handleMessageAsync<Messages::RemoteCDMInstanceProxy::SetServerCertificate>(connection, decoder, this, &RemoteCDMInstanceProxy::setServerCertificate);
        return;
    }
    if (decoder.messageName() == Messages::RemoteCDMInstanceProxy::SetStorageDirectory::name()) {
        IPC::handleMessage<Messages::RemoteCDMInstanceProxy::SetStorageDirectory>(decoder, this, &RemoteCDMInstanceProxy::setStorageDirectory);
        return;
    }
    UNUSED_PARAM(connection);
    UNUSED_PARAM(decoder);
    ASSERT_NOT_REACHED();
}

void RemoteCDMInstanceProxy::didReceiveSyncMessage(IPC::Connection& connection, IPC::Decoder& decoder, std::unique_ptr<IPC::Encoder>& replyEncoder)
{
    if (decoder.messageName() == Messages::RemoteCDMInstanceProxy::CreateSession::name()) {
        IPC::handleMessageSynchronous<Messages::RemoteCDMInstanceProxy::CreateSession>(connection, decoder, replyEncoder, this, &RemoteCDMInstanceProxy::createSession);
        return;
    }
    UNUSED_PARAM(connection);
    UNUSED_PARAM(decoder);
    UNUSED_PARAM(replyEncoder);
    ASSERT_NOT_REACHED();
}

} // namespace WebKit

#endif // ENABLE(GPU_PROCESS) && ENABLE(ENCRYPTED_MEDIA)
