/*
 * Copyright (C) 2010-2021 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#if ENABLE(GPU_PROCESS)
#include "RemoteMediaPlayerProxy.h"

#include "ArgumentCoders.h"
#include "Decoder.h"
#include "HandleMessage.h"
#include "LayerHostingContext.h"
#if ENABLE(ENCRYPTED_MEDIA)
#include "RemoteCDMInstanceIdentifier.h"
#endif
#if ENABLE(LEGACY_ENCRYPTED_MEDIA)
#include "RemoteLegacyCDMSessionIdentifier.h"
#endif
#include "RemoteMediaPlayerConfiguration.h"
#include "RemoteMediaPlayerProxyMessages.h"
#if ENABLE(MEDIA_SOURCE)
#include "RemoteMediaSourceIdentifier.h"
#endif
#include "SandboxExtension.h"
#include "TrackPrivateRemoteIdentifier.h"
#include "WebCoreArgumentCoders.h"
#include <WebCore/ContentType.h>
#include <WebCore/InbandTextTrackPrivate.h>
#if PLATFORM(COCOA)
#include <WebCore/IntSize.h>
#endif
#if ENABLE(WIRELESS_PLAYBACK_TARGET)
#include <WebCore/MediaPlaybackTargetContext.h>
#endif
#include <WebCore/MediaPlayer.h>
#include <WebCore/MediaPlayerEnums.h>
#include <WebCore/SecurityOriginData.h>
#if PLATFORM(COCOA)
#include <wtf/MachSendRight.h>
#endif
#include <wtf/MediaTime.h>
#include <wtf/Optional.h>
#include <wtf/WallTime.h>
#include <wtf/text/WTFString.h>

namespace Messages {

namespace RemoteMediaPlayerProxy {

void PrepareForPlayback::callReply(IPC::Decoder& decoder, CompletionHandler<void(Optional<WebKit::LayerHostingContextID>&&)>&& completionHandler)
{
    Optional<Optional<WebKit::LayerHostingContextID>> inlineLayerHostingContextId;
    decoder >> inlineLayerHostingContextId;
    if (!inlineLayerHostingContextId) {
        ASSERT_NOT_REACHED();
        cancelReply(WTFMove(completionHandler));
        return;
    }
    completionHandler(WTFMove(*inlineLayerHostingContextId));
}

void PrepareForPlayback::cancelReply(CompletionHandler<void(Optional<WebKit::LayerHostingContextID>&&)>&& completionHandler)
{
    completionHandler(IPC::AsyncReplyError<Optional<WebKit::LayerHostingContextID>>::create());
}

void PrepareForPlayback::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, const Optional<WebKit::LayerHostingContextID>& inlineLayerHostingContextId)
{
    *encoder << inlineLayerHostingContextId;
    connection.sendSyncReply(WTFMove(encoder));
}

void Load::callReply(IPC::Decoder& decoder, CompletionHandler<void(WebKit::RemoteMediaPlayerConfiguration&&)>&& completionHandler)
{
    Optional<WebKit::RemoteMediaPlayerConfiguration> playerConfiguration;
    decoder >> playerConfiguration;
    if (!playerConfiguration) {
        ASSERT_NOT_REACHED();
        cancelReply(WTFMove(completionHandler));
        return;
    }
    completionHandler(WTFMove(*playerConfiguration));
}

void Load::cancelReply(CompletionHandler<void(WebKit::RemoteMediaPlayerConfiguration&&)>&& completionHandler)
{
    completionHandler(IPC::AsyncReplyError<WebKit::RemoteMediaPlayerConfiguration>::create());
}

void Load::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, const WebKit::RemoteMediaPlayerConfiguration& playerConfiguration)
{
    *encoder << playerConfiguration;
    connection.sendSyncReply(WTFMove(encoder));
}

#if ENABLE(MEDIA_SOURCE)

void LoadMediaSource::callReply(IPC::Decoder& decoder, CompletionHandler<void(WebKit::RemoteMediaPlayerConfiguration&&)>&& completionHandler)
{
    Optional<WebKit::RemoteMediaPlayerConfiguration> playerConfiguration;
    decoder >> playerConfiguration;
    if (!playerConfiguration) {
        ASSERT_NOT_REACHED();
        cancelReply(WTFMove(completionHandler));
        return;
    }
    completionHandler(WTFMove(*playerConfiguration));
}

void LoadMediaSource::cancelReply(CompletionHandler<void(WebKit::RemoteMediaPlayerConfiguration&&)>&& completionHandler)
{
    completionHandler(IPC::AsyncReplyError<WebKit::RemoteMediaPlayerConfiguration>::create());
}

void LoadMediaSource::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, const WebKit::RemoteMediaPlayerConfiguration& playerConfiguration)
{
    *encoder << playerConfiguration;
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

void PerformTaskAtMediaTime::callReply(IPC::Decoder& decoder, CompletionHandler<void(Optional<MediaTime>&&)>&& completionHandler)
{
    Optional<Optional<MediaTime>> mediaTime;
    decoder >> mediaTime;
    if (!mediaTime) {
        ASSERT_NOT_REACHED();
        cancelReply(WTFMove(completionHandler));
        return;
    }
    completionHandler(WTFMove(*mediaTime));
}

void PerformTaskAtMediaTime::cancelReply(CompletionHandler<void(Optional<MediaTime>&&)>&& completionHandler)
{
    completionHandler(IPC::AsyncReplyError<Optional<MediaTime>>::create());
}

void PerformTaskAtMediaTime::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, const Optional<MediaTime>& mediaTime)
{
    *encoder << mediaTime;
    connection.sendSyncReply(WTFMove(encoder));
}

void WouldTaintOrigin::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, const Optional<bool>& wouldTaint)
{
    *encoder << wouldTaint;
    connection.sendSyncReply(WTFMove(encoder));
}

#if PLATFORM(IOS_FAMILY)

void ErrorLog::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, const String& errorLog)
{
    *encoder << errorLog;
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

#if PLATFORM(IOS_FAMILY)

void AccessLog::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, const String& accessLog)
{
    *encoder << accessLog;
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

} // namespace RemoteMediaPlayerProxy

} // namespace Messages

namespace WebKit {

void RemoteMediaPlayerProxy::didReceiveMessage(IPC::Connection& connection, IPC::Decoder& decoder)
{
    if (decoder.messageName() == Messages::RemoteMediaPlayerProxy::PrepareForPlayback::name()) {
        IPC::handleMessageAsync<Messages::RemoteMediaPlayerProxy::PrepareForPlayback>(connection, decoder, this, &RemoteMediaPlayerProxy::prepareForPlayback);
        return;
    }
    if (decoder.messageName() == Messages::RemoteMediaPlayerProxy::Load::name()) {
        IPC::handleMessageAsync<Messages::RemoteMediaPlayerProxy::Load>(connection, decoder, this, &RemoteMediaPlayerProxy::load);
        return;
    }
#if ENABLE(MEDIA_SOURCE)
    if (decoder.messageName() == Messages::RemoteMediaPlayerProxy::LoadMediaSource::name()) {
        IPC::handleMessageAsync<Messages::RemoteMediaPlayerProxy::LoadMediaSource>(connection, decoder, this, &RemoteMediaPlayerProxy::loadMediaSource);
        return;
    }
#endif
    if (decoder.messageName() == Messages::RemoteMediaPlayerProxy::CancelLoad::name()) {
        IPC::handleMessage<Messages::RemoteMediaPlayerProxy::CancelLoad>(decoder, this, &RemoteMediaPlayerProxy::cancelLoad);
        return;
    }
    if (decoder.messageName() == Messages::RemoteMediaPlayerProxy::PrepareToPlay::name()) {
        IPC::handleMessage<Messages::RemoteMediaPlayerProxy::PrepareToPlay>(decoder, this, &RemoteMediaPlayerProxy::prepareToPlay);
        return;
    }
    if (decoder.messageName() == Messages::RemoteMediaPlayerProxy::Play::name()) {
        IPC::handleMessage<Messages::RemoteMediaPlayerProxy::Play>(decoder, this, &RemoteMediaPlayerProxy::play);
        return;
    }
    if (decoder.messageName() == Messages::RemoteMediaPlayerProxy::Pause::name()) {
        IPC::handleMessage<Messages::RemoteMediaPlayerProxy::Pause>(decoder, this, &RemoteMediaPlayerProxy::pause);
        return;
    }
    if (decoder.messageName() == Messages::RemoteMediaPlayerProxy::SetVolume::name()) {
        IPC::handleMessage<Messages::RemoteMediaPlayerProxy::SetVolume>(decoder, this, &RemoteMediaPlayerProxy::setVolume);
        return;
    }
    if (decoder.messageName() == Messages::RemoteMediaPlayerProxy::SetMuted::name()) {
        IPC::handleMessage<Messages::RemoteMediaPlayerProxy::SetMuted>(decoder, this, &RemoteMediaPlayerProxy::setMuted);
        return;
    }
    if (decoder.messageName() == Messages::RemoteMediaPlayerProxy::Seek::name()) {
        IPC::handleMessage<Messages::RemoteMediaPlayerProxy::Seek>(decoder, this, &RemoteMediaPlayerProxy::seek);
        return;
    }
    if (decoder.messageName() == Messages::RemoteMediaPlayerProxy::SeekWithTolerance::name()) {
        IPC::handleMessage<Messages::RemoteMediaPlayerProxy::SeekWithTolerance>(decoder, this, &RemoteMediaPlayerProxy::seekWithTolerance);
        return;
    }
    if (decoder.messageName() == Messages::RemoteMediaPlayerProxy::SetPreload::name()) {
        IPC::handleMessage<Messages::RemoteMediaPlayerProxy::SetPreload>(decoder, this, &RemoteMediaPlayerProxy::setPreload);
        return;
    }
    if (decoder.messageName() == Messages::RemoteMediaPlayerProxy::SetPrivateBrowsingMode::name()) {
        IPC::handleMessage<Messages::RemoteMediaPlayerProxy::SetPrivateBrowsingMode>(decoder, this, &RemoteMediaPlayerProxy::setPrivateBrowsingMode);
        return;
    }
    if (decoder.messageName() == Messages::RemoteMediaPlayerProxy::SetPreservesPitch::name()) {
        IPC::handleMessage<Messages::RemoteMediaPlayerProxy::SetPreservesPitch>(decoder, this, &RemoteMediaPlayerProxy::setPreservesPitch);
        return;
    }
    if (decoder.messageName() == Messages::RemoteMediaPlayerProxy::SetPitchCorrectionAlgorithm::name()) {
        IPC::handleMessage<Messages::RemoteMediaPlayerProxy::SetPitchCorrectionAlgorithm>(decoder, this, &RemoteMediaPlayerProxy::setPitchCorrectionAlgorithm);
        return;
    }
    if (decoder.messageName() == Messages::RemoteMediaPlayerProxy::PrepareForRendering::name()) {
        IPC::handleMessage<Messages::RemoteMediaPlayerProxy::PrepareForRendering>(decoder, this, &RemoteMediaPlayerProxy::prepareForRendering);
        return;
    }
    if (decoder.messageName() == Messages::RemoteMediaPlayerProxy::SetVisible::name()) {
        IPC::handleMessage<Messages::RemoteMediaPlayerProxy::SetVisible>(decoder, this, &RemoteMediaPlayerProxy::setVisible);
        return;
    }
    if (decoder.messageName() == Messages::RemoteMediaPlayerProxy::SetShouldMaintainAspectRatio::name()) {
        IPC::handleMessage<Messages::RemoteMediaPlayerProxy::SetShouldMaintainAspectRatio>(decoder, this, &RemoteMediaPlayerProxy::setShouldMaintainAspectRatio);
        return;
    }
#if ENABLE(VIDEO_PRESENTATION_MODE)
    if (decoder.messageName() == Messages::RemoteMediaPlayerProxy::SetVideoFullscreenGravity::name()) {
        IPC::handleMessage<Messages::RemoteMediaPlayerProxy::SetVideoFullscreenGravity>(decoder, this, &RemoteMediaPlayerProxy::setVideoFullscreenGravity);
        return;
    }
#endif
    if (decoder.messageName() == Messages::RemoteMediaPlayerProxy::AcceleratedRenderingStateChanged::name()) {
        IPC::handleMessage<Messages::RemoteMediaPlayerProxy::AcceleratedRenderingStateChanged>(decoder, this, &RemoteMediaPlayerProxy::acceleratedRenderingStateChanged);
        return;
    }
    if (decoder.messageName() == Messages::RemoteMediaPlayerProxy::SetShouldDisableSleep::name()) {
        IPC::handleMessage<Messages::RemoteMediaPlayerProxy::SetShouldDisableSleep>(decoder, this, &RemoteMediaPlayerProxy::setShouldDisableSleep);
        return;
    }
    if (decoder.messageName() == Messages::RemoteMediaPlayerProxy::SetRate::name()) {
        IPC::handleMessage<Messages::RemoteMediaPlayerProxy::SetRate>(decoder, this, &RemoteMediaPlayerProxy::setRate);
        return;
    }
#if ENABLE(VIDEO_PRESENTATION_MODE)
    if (decoder.messageName() == Messages::RemoteMediaPlayerProxy::UpdateVideoFullscreenInlineImage::name()) {
        IPC::handleMessage<Messages::RemoteMediaPlayerProxy::UpdateVideoFullscreenInlineImage>(decoder, this, &RemoteMediaPlayerProxy::updateVideoFullscreenInlineImage);
        return;
    }
#endif
#if ENABLE(VIDEO_PRESENTATION_MODE)
    if (decoder.messageName() == Messages::RemoteMediaPlayerProxy::SetVideoFullscreenMode::name()) {
        IPC::handleMessage<Messages::RemoteMediaPlayerProxy::SetVideoFullscreenMode>(decoder, this, &RemoteMediaPlayerProxy::setVideoFullscreenMode);
        return;
    }
#endif
#if ENABLE(VIDEO_PRESENTATION_MODE)
    if (decoder.messageName() == Messages::RemoteMediaPlayerProxy::VideoFullscreenStandbyChanged::name()) {
        IPC::handleMessage<Messages::RemoteMediaPlayerProxy::VideoFullscreenStandbyChanged>(decoder, this, &RemoteMediaPlayerProxy::videoFullscreenStandbyChanged);
        return;
    }
#endif
    if (decoder.messageName() == Messages::RemoteMediaPlayerProxy::SetBufferingPolicy::name()) {
        IPC::handleMessage<Messages::RemoteMediaPlayerProxy::SetBufferingPolicy>(decoder, this, &RemoteMediaPlayerProxy::setBufferingPolicy);
        return;
    }
#if PLATFORM(COCOA)
    if (decoder.messageName() == Messages::RemoteMediaPlayerProxy::SetVideoInlineSizeFenced::name()) {
        IPC::handleMessage<Messages::RemoteMediaPlayerProxy::SetVideoInlineSizeFenced>(decoder, this, &RemoteMediaPlayerProxy::setVideoInlineSizeFenced);
        return;
    }
#endif
#if ENABLE(WIRELESS_PLAYBACK_TARGET)
    if (decoder.messageName() == Messages::RemoteMediaPlayerProxy::SetWirelessVideoPlaybackDisabled::name()) {
        IPC::handleMessage<Messages::RemoteMediaPlayerProxy::SetWirelessVideoPlaybackDisabled>(decoder, this, &RemoteMediaPlayerProxy::setWirelessVideoPlaybackDisabled);
        return;
    }
#endif
#if ENABLE(WIRELESS_PLAYBACK_TARGET)
    if (decoder.messageName() == Messages::RemoteMediaPlayerProxy::SetShouldPlayToPlaybackTarget::name()) {
        IPC::handleMessage<Messages::RemoteMediaPlayerProxy::SetShouldPlayToPlaybackTarget>(decoder, this, &RemoteMediaPlayerProxy::setShouldPlayToPlaybackTarget);
        return;
    }
#endif
#if ENABLE(LEGACY_ENCRYPTED_MEDIA)
    if (decoder.messageName() == Messages::RemoteMediaPlayerProxy::SetLegacyCDMSession::name()) {
        IPC::handleMessage<Messages::RemoteMediaPlayerProxy::SetLegacyCDMSession>(decoder, this, &RemoteMediaPlayerProxy::setLegacyCDMSession);
        return;
    }
#endif
#if ENABLE(LEGACY_ENCRYPTED_MEDIA)
    if (decoder.messageName() == Messages::RemoteMediaPlayerProxy::KeyAdded::name()) {
        IPC::handleMessage<Messages::RemoteMediaPlayerProxy::KeyAdded>(decoder, this, &RemoteMediaPlayerProxy::keyAdded);
        return;
    }
#endif
#if ENABLE(ENCRYPTED_MEDIA)
    if (decoder.messageName() == Messages::RemoteMediaPlayerProxy::CdmInstanceAttached::name()) {
        IPC::handleMessage<Messages::RemoteMediaPlayerProxy::CdmInstanceAttached>(decoder, this, &RemoteMediaPlayerProxy::cdmInstanceAttached);
        return;
    }
#endif
#if ENABLE(ENCRYPTED_MEDIA)
    if (decoder.messageName() == Messages::RemoteMediaPlayerProxy::CdmInstanceDetached::name()) {
        IPC::handleMessage<Messages::RemoteMediaPlayerProxy::CdmInstanceDetached>(decoder, this, &RemoteMediaPlayerProxy::cdmInstanceDetached);
        return;
    }
#endif
#if ENABLE(ENCRYPTED_MEDIA)
    if (decoder.messageName() == Messages::RemoteMediaPlayerProxy::AttemptToDecryptWithInstance::name()) {
        IPC::handleMessage<Messages::RemoteMediaPlayerProxy::AttemptToDecryptWithInstance>(decoder, this, &RemoteMediaPlayerProxy::attemptToDecryptWithInstance);
        return;
    }
#endif
#if ENABLE(LEGACY_ENCRYPTED_MEDIA) && ENABLE(ENCRYPTED_MEDIA)
    if (decoder.messageName() == Messages::RemoteMediaPlayerProxy::SetShouldContinueAfterKeyNeeded::name()) {
        IPC::handleMessage<Messages::RemoteMediaPlayerProxy::SetShouldContinueAfterKeyNeeded>(decoder, this, &RemoteMediaPlayerProxy::setShouldContinueAfterKeyNeeded);
        return;
    }
#endif
    if (decoder.messageName() == Messages::RemoteMediaPlayerProxy::BeginSimulatedHDCPError::name()) {
        IPC::handleMessage<Messages::RemoteMediaPlayerProxy::BeginSimulatedHDCPError>(decoder, this, &RemoteMediaPlayerProxy::beginSimulatedHDCPError);
        return;
    }
    if (decoder.messageName() == Messages::RemoteMediaPlayerProxy::EndSimulatedHDCPError::name()) {
        IPC::handleMessage<Messages::RemoteMediaPlayerProxy::EndSimulatedHDCPError>(decoder, this, &RemoteMediaPlayerProxy::endSimulatedHDCPError);
        return;
    }
    if (decoder.messageName() == Messages::RemoteMediaPlayerProxy::NotifyActiveSourceBuffersChanged::name()) {
        IPC::handleMessage<Messages::RemoteMediaPlayerProxy::NotifyActiveSourceBuffersChanged>(decoder, this, &RemoteMediaPlayerProxy::notifyActiveSourceBuffersChanged);
        return;
    }
    if (decoder.messageName() == Messages::RemoteMediaPlayerProxy::ApplicationWillResignActive::name()) {
        IPC::handleMessage<Messages::RemoteMediaPlayerProxy::ApplicationWillResignActive>(decoder, this, &RemoteMediaPlayerProxy::applicationWillResignActive);
        return;
    }
    if (decoder.messageName() == Messages::RemoteMediaPlayerProxy::ApplicationDidBecomeActive::name()) {
        IPC::handleMessage<Messages::RemoteMediaPlayerProxy::ApplicationDidBecomeActive>(decoder, this, &RemoteMediaPlayerProxy::applicationDidBecomeActive);
        return;
    }
    if (decoder.messageName() == Messages::RemoteMediaPlayerProxy::NotifyTrackModeChanged::name()) {
        IPC::handleMessage<Messages::RemoteMediaPlayerProxy::NotifyTrackModeChanged>(decoder, this, &RemoteMediaPlayerProxy::notifyTrackModeChanged);
        return;
    }
    if (decoder.messageName() == Messages::RemoteMediaPlayerProxy::TracksChanged::name()) {
        IPC::handleMessage<Messages::RemoteMediaPlayerProxy::TracksChanged>(decoder, this, &RemoteMediaPlayerProxy::tracksChanged);
        return;
    }
    if (decoder.messageName() == Messages::RemoteMediaPlayerProxy::AudioTrackSetEnabled::name()) {
        IPC::handleMessage<Messages::RemoteMediaPlayerProxy::AudioTrackSetEnabled>(decoder, this, &RemoteMediaPlayerProxy::audioTrackSetEnabled);
        return;
    }
    if (decoder.messageName() == Messages::RemoteMediaPlayerProxy::VideoTrackSetSelected::name()) {
        IPC::handleMessage<Messages::RemoteMediaPlayerProxy::VideoTrackSetSelected>(decoder, this, &RemoteMediaPlayerProxy::videoTrackSetSelected);
        return;
    }
    if (decoder.messageName() == Messages::RemoteMediaPlayerProxy::TextTrackSetMode::name()) {
        IPC::handleMessage<Messages::RemoteMediaPlayerProxy::TextTrackSetMode>(decoder, this, &RemoteMediaPlayerProxy::textTrackSetMode);
        return;
    }
#if ENABLE(WIRELESS_PLAYBACK_TARGET)
    if (decoder.messageName() == Messages::RemoteMediaPlayerProxy::SetWirelessPlaybackTarget::name()) {
        IPC::handleMessage<Messages::RemoteMediaPlayerProxy::SetWirelessPlaybackTarget>(decoder, this, &RemoteMediaPlayerProxy::setWirelessPlaybackTarget);
        return;
    }
#endif
    if (decoder.messageName() == Messages::RemoteMediaPlayerProxy::PerformTaskAtMediaTime::name()) {
        IPC::handleMessageAsync<Messages::RemoteMediaPlayerProxy::PerformTaskAtMediaTime>(connection, decoder, this, &RemoteMediaPlayerProxy::performTaskAtMediaTime);
        return;
    }
    if (decoder.messageName() == Messages::RemoteMediaPlayerProxy::SetShouldUpdatePlaybackMetrics::name()) {
        IPC::handleMessage<Messages::RemoteMediaPlayerProxy::SetShouldUpdatePlaybackMetrics>(decoder, this, &RemoteMediaPlayerProxy::setShouldUpdatePlaybackMetrics);
        return;
    }
#if ENABLE(WEB_AUDIO)
    if (decoder.messageName() == Messages::RemoteMediaPlayerProxy::CreateAudioSourceProvider::name()) {
        IPC::handleMessage<Messages::RemoteMediaPlayerProxy::CreateAudioSourceProvider>(decoder, this, &RemoteMediaPlayerProxy::createAudioSourceProvider);
        return;
    }
#endif
#if ENABLE(WEB_AUDIO)
    if (decoder.messageName() == Messages::RemoteMediaPlayerProxy::SetShouldEnableAudioSourceProvider::name()) {
        IPC::handleMessage<Messages::RemoteMediaPlayerProxy::SetShouldEnableAudioSourceProvider>(decoder, this, &RemoteMediaPlayerProxy::setShouldEnableAudioSourceProvider);
        return;
    }
#endif
    UNUSED_PARAM(connection);
    UNUSED_PARAM(decoder);
    ASSERT_NOT_REACHED();
}

void RemoteMediaPlayerProxy::didReceiveSyncMessage(IPC::Connection& connection, IPC::Decoder& decoder, std::unique_ptr<IPC::Encoder>& replyEncoder)
{
    if (decoder.messageName() == Messages::RemoteMediaPlayerProxy::WouldTaintOrigin::name()) {
        IPC::handleMessageSynchronous<Messages::RemoteMediaPlayerProxy::WouldTaintOrigin>(connection, decoder, replyEncoder, this, &RemoteMediaPlayerProxy::wouldTaintOrigin);
        return;
    }
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::RemoteMediaPlayerProxy::ErrorLog::name()) {
        IPC::handleMessageSynchronous<Messages::RemoteMediaPlayerProxy::ErrorLog>(connection, decoder, replyEncoder, this, &RemoteMediaPlayerProxy::errorLog);
        return;
    }
#endif
#if PLATFORM(IOS_FAMILY)
    if (decoder.messageName() == Messages::RemoteMediaPlayerProxy::AccessLog::name()) {
        IPC::handleMessageSynchronous<Messages::RemoteMediaPlayerProxy::AccessLog>(connection, decoder, replyEncoder, this, &RemoteMediaPlayerProxy::accessLog);
        return;
    }
#endif
    UNUSED_PARAM(connection);
    UNUSED_PARAM(decoder);
    UNUSED_PARAM(replyEncoder);
    ASSERT_NOT_REACHED();
}

} // namespace WebKit

#endif // ENABLE(GPU_PROCESS)
