/*
 * Copyright (C) 2010-2021 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#if ENABLE(GPU_PROCESS)
#include "RemoteRenderingBackend.h"

#include "ArgumentCoders.h"
#include "Decoder.h"
#include "FontReference.h"
#include "GPUProcessWakeupMessageArguments.h"
#include "HandleMessage.h"
#include "RemoteRenderingBackendMessages.h"
#include "ShareableBitmap.h"
#include "SharedMemory.h"
#include "WebCoreArgumentCoders.h"
#include <WebCore/AlphaPremultiplication.h>
#include <WebCore/ColorSpace.h>
#include <WebCore/DisplayList.h>
#include <WebCore/FloatSize.h>
#include <WebCore/ImageBufferBackend.h>
#include <WebCore/ImageData.h>
#include <WebCore/IntRect.h>
#include <WebCore/RenderingMode.h>
#include <WebCore/RenderingResourceIdentifier.h>
#include <wtf/Optional.h>
#include <wtf/RefCounted.h>
#include <wtf/Vector.h>
#include <wtf/text/WTFString.h>

namespace Messages {

namespace RemoteRenderingBackend {

void GetImageData::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, const RefPtr<WebCore::ImageData>& imageData)
{
    *encoder << imageData;
    connection.sendSyncReply(WTFMove(encoder));
}

void GetDataURLForImageBuffer::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, const String& urlString)
{
    *encoder << urlString;
    connection.sendSyncReply(WTFMove(encoder));
}

void GetDataForImageBuffer::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, const Vector<uint8_t>& data)
{
    *encoder << data;
    connection.sendSyncReply(WTFMove(encoder));
}

void GetBGRADataForImageBuffer::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, const Vector<uint8_t>& data)
{
    *encoder << data;
    connection.sendSyncReply(WTFMove(encoder));
}

void GetShareableBitmapForImageBuffer::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, const WebKit::ShareableBitmap::Handle& handle)
{
    *encoder << handle;
    connection.sendSyncReply(WTFMove(encoder));
}

} // namespace RemoteRenderingBackend

} // namespace Messages

namespace WebKit {

void RemoteRenderingBackend::didReceiveMessage(IPC::Connection& connection, IPC::Decoder& decoder)
{
    auto protectedThis = makeRef(*this);
    if (decoder.messageName() == Messages::RemoteRenderingBackend::CreateImageBuffer::name()) {
        IPC::handleMessage<Messages::RemoteRenderingBackend::CreateImageBuffer>(decoder, this, &RemoteRenderingBackend::createImageBuffer);
        return;
    }
    if (decoder.messageName() == Messages::RemoteRenderingBackend::WakeUpAndApplyDisplayList::name()) {
        IPC::handleMessage<Messages::RemoteRenderingBackend::WakeUpAndApplyDisplayList>(decoder, this, &RemoteRenderingBackend::wakeUpAndApplyDisplayList);
        return;
    }
    if (decoder.messageName() == Messages::RemoteRenderingBackend::CacheNativeImage::name()) {
        IPC::handleMessage<Messages::RemoteRenderingBackend::CacheNativeImage>(decoder, this, &RemoteRenderingBackend::cacheNativeImage);
        return;
    }
    if (decoder.messageName() == Messages::RemoteRenderingBackend::CacheFont::name()) {
        IPC::handleMessage<Messages::RemoteRenderingBackend::CacheFont>(decoder, this, &RemoteRenderingBackend::cacheFont);
        return;
    }
    if (decoder.messageName() == Messages::RemoteRenderingBackend::DeleteAllFonts::name()) {
        IPC::handleMessage<Messages::RemoteRenderingBackend::DeleteAllFonts>(decoder, this, &RemoteRenderingBackend::deleteAllFonts);
        return;
    }
    if (decoder.messageName() == Messages::RemoteRenderingBackend::DidCreateSharedDisplayListHandle::name()) {
        IPC::handleMessage<Messages::RemoteRenderingBackend::DidCreateSharedDisplayListHandle>(decoder, this, &RemoteRenderingBackend::didCreateSharedDisplayListHandle);
        return;
    }
    if (decoder.messageName() == Messages::RemoteRenderingBackend::ReleaseRemoteResource::name()) {
        IPC::handleMessage<Messages::RemoteRenderingBackend::ReleaseRemoteResource>(decoder, this, &RemoteRenderingBackend::releaseRemoteResource);
        return;
    }
    UNUSED_PARAM(connection);
    UNUSED_PARAM(decoder);
    ASSERT_NOT_REACHED();
}

void RemoteRenderingBackend::didReceiveSyncMessage(IPC::Connection& connection, IPC::Decoder& decoder, std::unique_ptr<IPC::Encoder>& replyEncoder)
{
    auto protectedThis = makeRef(*this);
    if (decoder.messageName() == Messages::RemoteRenderingBackend::GetImageData::name()) {
        IPC::handleMessageSynchronous<Messages::RemoteRenderingBackend::GetImageData>(connection, decoder, replyEncoder, this, &RemoteRenderingBackend::getImageData);
        return;
    }
    if (decoder.messageName() == Messages::RemoteRenderingBackend::GetDataURLForImageBuffer::name()) {
        IPC::handleMessageSynchronous<Messages::RemoteRenderingBackend::GetDataURLForImageBuffer>(connection, decoder, replyEncoder, this, &RemoteRenderingBackend::getDataURLForImageBuffer);
        return;
    }
    if (decoder.messageName() == Messages::RemoteRenderingBackend::GetDataForImageBuffer::name()) {
        IPC::handleMessageSynchronous<Messages::RemoteRenderingBackend::GetDataForImageBuffer>(connection, decoder, replyEncoder, this, &RemoteRenderingBackend::getDataForImageBuffer);
        return;
    }
    if (decoder.messageName() == Messages::RemoteRenderingBackend::GetBGRADataForImageBuffer::name()) {
        IPC::handleMessageSynchronous<Messages::RemoteRenderingBackend::GetBGRADataForImageBuffer>(connection, decoder, replyEncoder, this, &RemoteRenderingBackend::getBGRADataForImageBuffer);
        return;
    }
    if (decoder.messageName() == Messages::RemoteRenderingBackend::GetShareableBitmapForImageBuffer::name()) {
        IPC::handleMessageSynchronous<Messages::RemoteRenderingBackend::GetShareableBitmapForImageBuffer>(connection, decoder, replyEncoder, this, &RemoteRenderingBackend::getShareableBitmapForImageBuffer);
        return;
    }
    UNUSED_PARAM(connection);
    UNUSED_PARAM(decoder);
    UNUSED_PARAM(replyEncoder);
    ASSERT_NOT_REACHED();
}

} // namespace WebKit

#endif // ENABLE(GPU_PROCESS)
