/*
 * Copyright (C) 2010-2021 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#if ENABLE(GPU_PROCESS) && ENABLE(MEDIA_SOURCE)
#include "RemoteSourceBufferProxy.h"

#include "ArgumentCoders.h"
#include "DataReference.h"
#include "Decoder.h"
#include "HandleMessage.h"
#include "RemoteSourceBufferProxyMessages.h"
#include "TrackPrivateRemoteIdentifier.h"
#include "WebCoreArgumentCoders.h"
#include <WebCore/MediaPlayer.h>
#include <WebCore/SourceBufferPrivate.h>
#include <utility>
#include <wtf/MediaTime.h>
#include <wtf/Vector.h>
#include <wtf/text/WTFString.h>

namespace Messages {

namespace RemoteSourceBufferProxy {

void RemoveCodedFrames::callReply(IPC::Decoder& decoder, CompletionHandler<void()>&& completionHandler)
{
    completionHandler();
}

void RemoveCodedFrames::cancelReply(CompletionHandler<void()>&& completionHandler)
{
    completionHandler();
}

void RemoveCodedFrames::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection)
{
    connection.sendSyncReply(WTFMove(encoder));
}

void BufferedSamplesForTrackId::callReply(IPC::Decoder& decoder, CompletionHandler<void(Vector<String>&&)>&& completionHandler)
{
    Optional<Vector<String>> samples;
    decoder >> samples;
    if (!samples) {
        ASSERT_NOT_REACHED();
        cancelReply(WTFMove(completionHandler));
        return;
    }
    completionHandler(WTFMove(*samples));
}

void BufferedSamplesForTrackId::cancelReply(CompletionHandler<void(Vector<String>&&)>&& completionHandler)
{
    completionHandler(IPC::AsyncReplyError<Vector<String>>::create());
}

void BufferedSamplesForTrackId::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, const Vector<String>& samples)
{
    *encoder << samples;
    connection.sendSyncReply(WTFMove(encoder));
}

} // namespace RemoteSourceBufferProxy

} // namespace Messages

namespace WebKit {

void RemoteSourceBufferProxy::didReceiveMessage(IPC::Connection& connection, IPC::Decoder& decoder)
{
    if (decoder.messageName() == Messages::RemoteSourceBufferProxy::SetActive::name()) {
        IPC::handleMessage<Messages::RemoteSourceBufferProxy::SetActive>(decoder, this, &RemoteSourceBufferProxy::setActive);
        return;
    }
    if (decoder.messageName() == Messages::RemoteSourceBufferProxy::SetMode::name()) {
        IPC::handleMessage<Messages::RemoteSourceBufferProxy::SetMode>(decoder, this, &RemoteSourceBufferProxy::setMode);
        return;
    }
    if (decoder.messageName() == Messages::RemoteSourceBufferProxy::Append::name()) {
        IPC::handleMessage<Messages::RemoteSourceBufferProxy::Append>(decoder, this, &RemoteSourceBufferProxy::append);
        return;
    }
    if (decoder.messageName() == Messages::RemoteSourceBufferProxy::Abort::name()) {
        IPC::handleMessage<Messages::RemoteSourceBufferProxy::Abort>(decoder, this, &RemoteSourceBufferProxy::abort);
        return;
    }
    if (decoder.messageName() == Messages::RemoteSourceBufferProxy::ResetParserState::name()) {
        IPC::handleMessage<Messages::RemoteSourceBufferProxy::ResetParserState>(decoder, this, &RemoteSourceBufferProxy::resetParserState);
        return;
    }
    if (decoder.messageName() == Messages::RemoteSourceBufferProxy::RemovedFromMediaSource::name()) {
        IPC::handleMessage<Messages::RemoteSourceBufferProxy::RemovedFromMediaSource>(decoder, this, &RemoteSourceBufferProxy::removedFromMediaSource);
        return;
    }
    if (decoder.messageName() == Messages::RemoteSourceBufferProxy::SetMediaSourceEnded::name()) {
        IPC::handleMessage<Messages::RemoteSourceBufferProxy::SetMediaSourceEnded>(decoder, this, &RemoteSourceBufferProxy::setMediaSourceEnded);
        return;
    }
    if (decoder.messageName() == Messages::RemoteSourceBufferProxy::SetReadyState::name()) {
        IPC::handleMessage<Messages::RemoteSourceBufferProxy::SetReadyState>(decoder, this, &RemoteSourceBufferProxy::setReadyState);
        return;
    }
    if (decoder.messageName() == Messages::RemoteSourceBufferProxy::StartChangingType::name()) {
        IPC::handleMessage<Messages::RemoteSourceBufferProxy::StartChangingType>(decoder, this, &RemoteSourceBufferProxy::startChangingType);
        return;
    }
    if (decoder.messageName() == Messages::RemoteSourceBufferProxy::UpdateBufferedFromTrackBuffers::name()) {
        IPC::handleMessage<Messages::RemoteSourceBufferProxy::UpdateBufferedFromTrackBuffers>(decoder, this, &RemoteSourceBufferProxy::updateBufferedFromTrackBuffers);
        return;
    }
    if (decoder.messageName() == Messages::RemoteSourceBufferProxy::AddTrackBuffer::name()) {
        IPC::handleMessage<Messages::RemoteSourceBufferProxy::AddTrackBuffer>(decoder, this, &RemoteSourceBufferProxy::addTrackBuffer);
        return;
    }
    if (decoder.messageName() == Messages::RemoteSourceBufferProxy::ResetTrackBuffers::name()) {
        IPC::handleMessage<Messages::RemoteSourceBufferProxy::ResetTrackBuffers>(decoder, this, &RemoteSourceBufferProxy::resetTrackBuffers);
        return;
    }
    if (decoder.messageName() == Messages::RemoteSourceBufferProxy::ClearTrackBuffers::name()) {
        IPC::handleMessage<Messages::RemoteSourceBufferProxy::ClearTrackBuffers>(decoder, this, &RemoteSourceBufferProxy::clearTrackBuffers);
        return;
    }
    if (decoder.messageName() == Messages::RemoteSourceBufferProxy::SetAllTrackBuffersNeedRandomAccess::name()) {
        IPC::handleMessage<Messages::RemoteSourceBufferProxy::SetAllTrackBuffersNeedRandomAccess>(decoder, this, &RemoteSourceBufferProxy::setAllTrackBuffersNeedRandomAccess);
        return;
    }
    if (decoder.messageName() == Messages::RemoteSourceBufferProxy::RemoveCodedFrames::name()) {
        IPC::handleMessageAsync<Messages::RemoteSourceBufferProxy::RemoveCodedFrames>(connection, decoder, this, &RemoteSourceBufferProxy::removeCodedFrames);
        return;
    }
    if (decoder.messageName() == Messages::RemoteSourceBufferProxy::EvictCodedFrames::name()) {
        IPC::handleMessage<Messages::RemoteSourceBufferProxy::EvictCodedFrames>(decoder, this, &RemoteSourceBufferProxy::evictCodedFrames);
        return;
    }
    if (decoder.messageName() == Messages::RemoteSourceBufferProxy::ReenqueueMediaIfNeeded::name()) {
        IPC::handleMessage<Messages::RemoteSourceBufferProxy::ReenqueueMediaIfNeeded>(decoder, this, &RemoteSourceBufferProxy::reenqueueMediaIfNeeded);
        return;
    }
    if (decoder.messageName() == Messages::RemoteSourceBufferProxy::SetGroupStartTimestamp::name()) {
        IPC::handleMessage<Messages::RemoteSourceBufferProxy::SetGroupStartTimestamp>(decoder, this, &RemoteSourceBufferProxy::setGroupStartTimestamp);
        return;
    }
    if (decoder.messageName() == Messages::RemoteSourceBufferProxy::SetGroupStartTimestampToEndTimestamp::name()) {
        IPC::handleMessage<Messages::RemoteSourceBufferProxy::SetGroupStartTimestampToEndTimestamp>(decoder, this, &RemoteSourceBufferProxy::setGroupStartTimestampToEndTimestamp);
        return;
    }
    if (decoder.messageName() == Messages::RemoteSourceBufferProxy::SetShouldGenerateTimestamps::name()) {
        IPC::handleMessage<Messages::RemoteSourceBufferProxy::SetShouldGenerateTimestamps>(decoder, this, &RemoteSourceBufferProxy::setShouldGenerateTimestamps);
        return;
    }
    if (decoder.messageName() == Messages::RemoteSourceBufferProxy::ResetTimestampOffsetInTrackBuffers::name()) {
        IPC::handleMessage<Messages::RemoteSourceBufferProxy::ResetTimestampOffsetInTrackBuffers>(decoder, this, &RemoteSourceBufferProxy::resetTimestampOffsetInTrackBuffers);
        return;
    }
    if (decoder.messageName() == Messages::RemoteSourceBufferProxy::SetTimestampOffset::name()) {
        IPC::handleMessage<Messages::RemoteSourceBufferProxy::SetTimestampOffset>(decoder, this, &RemoteSourceBufferProxy::setTimestampOffset);
        return;
    }
    if (decoder.messageName() == Messages::RemoteSourceBufferProxy::SetAppendWindowStart::name()) {
        IPC::handleMessage<Messages::RemoteSourceBufferProxy::SetAppendWindowStart>(decoder, this, &RemoteSourceBufferProxy::setAppendWindowStart);
        return;
    }
    if (decoder.messageName() == Messages::RemoteSourceBufferProxy::SetAppendWindowEnd::name()) {
        IPC::handleMessage<Messages::RemoteSourceBufferProxy::SetAppendWindowEnd>(decoder, this, &RemoteSourceBufferProxy::setAppendWindowEnd);
        return;
    }
    if (decoder.messageName() == Messages::RemoteSourceBufferProxy::SeekToTime::name()) {
        IPC::handleMessage<Messages::RemoteSourceBufferProxy::SeekToTime>(decoder, this, &RemoteSourceBufferProxy::seekToTime);
        return;
    }
    if (decoder.messageName() == Messages::RemoteSourceBufferProxy::UpdateTrackIds::name()) {
        IPC::handleMessage<Messages::RemoteSourceBufferProxy::UpdateTrackIds>(decoder, this, &RemoteSourceBufferProxy::updateTrackIds);
        return;
    }
    if (decoder.messageName() == Messages::RemoteSourceBufferProxy::BufferedSamplesForTrackId::name()) {
        IPC::handleMessageAsync<Messages::RemoteSourceBufferProxy::BufferedSamplesForTrackId>(connection, decoder, this, &RemoteSourceBufferProxy::bufferedSamplesForTrackId);
        return;
    }
    UNUSED_PARAM(connection);
    UNUSED_PARAM(decoder);
    ASSERT_NOT_REACHED();
}

} // namespace WebKit

#endif // ENABLE(GPU_PROCESS) && ENABLE(MEDIA_SOURCE)
